from enum import Enum
from dataclasses import dataclass, field
from typing import Optional, Any
from collections import namedtuple

@dataclass
class ResponseIds:
    """
    Used as a return value from the upsert operation on the DV.
    Contains the table names and ids of the entities that were 
    updated and created, stored separately.
    """
    def __init__(self) -> None:
        self.updated_entity_ids: dict[str, set[str]] = dict()
        self.created_entity_ids: dict[str, set[str]] = dict()
    
    
    def test(self):
        return "Welcome to OTCFinUtils"
    
    
    def append(self, status_code: int, id: str, table: str) -> None:
        """
        Depending on the status code appends the id and 
        table name to the create / update category.
        """
        if status_code == HTTPStatus.OK.value and table in self.updated_entity_ids:
            self.updated_entity_ids[table].add(id)
        
        elif status_code == HTTPStatus.OK.value and table not in self.updated_entity_ids:
            self.updated_entity_ids[table] = set()
            self.updated_entity_ids[table].add(id)
        
        elif status_code == HTTPStatus.CREATED.value and table in self.created_entity_ids:
            self.created_entity_ids[table].add(id)
        
        elif status_code == HTTPStatus.CREATED.value and table not in self.created_entity_ids:
            self.created_entity_ids[table] = set()
            self.created_entity_ids[table].add(id)
        
        return
    

    def merge(self, delta_ids: "ResponseIds") -> None:
        """
        Merges the values from the provided object onto the
        the values of the calling object.
        """
        for table, ids in delta_ids.created_entity_ids.items():
            if table not in self.created_entity_ids:
                self.created_entity_ids[table] = set()
            self.created_entity_ids[table].update(ids)

        for table, ids in delta_ids.updated_entity_ids.items():
            if table not in self.updated_entity_ids:
                self.updated_entity_ids[table] = set()
            self.updated_entity_ids[table].update(ids)

        return
    

    def json_updated_entities(self) -> dict[str, list]:
        """
        Used for getting a JSON compatible object.
        """
        result = dict()
        for table, ids in self.updated_entity_ids.items():
            result[table] = list(ids)
        return result
    

    def json_created_entities(self) -> dict[str, list]:
        """
        Used for getting a JSON compatible object.
        """
        result = dict()
        for table, ids in self.created_entity_ids.items():
            result[table] = list(ids)
        return result
    

    def __str__(self) -> str:
        return (
            f"created: {self.created_entity_ids}\n"
            f"updated: {self.updated_entity_ids}"
        )
    

@dataclass
class APIParams:
    """
    Basic helper data structure for making API calls.
    """
    url: str
    method: str
    data: dict | None = None


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    DELETE = "DELETE"


class HTTPStatus(Enum):
    OK = 200
    CREATED = 201
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    NOT_FOUND = 404
    INTERNAL_SERVER_ERROR = 500


@dataclass
class KeyLookupParams:
    lookup_id_column: str
    lookup_id: Optional[str]


EntityKeyLookupParams = dict[str, KeyLookupParams]


@dataclass
class IDParams:
    """
    A group of parameters used for querying id info of entities.
    """
    table: Optional[str] = None
    id_column: Optional[str] = None
    conditions: list[str] = field(default_factory=list)
    
    # If we have a lookup
    bind_column: Optional[str] = None
    

EntityLookupParams = list[IDParams]

"""
Used to divide and sort the file loading mappings
"""
Category = namedtuple("Category", ["group_num", "table", "sheet"])

"""
Used for storing the id info of the lookup entity
"""
Lookup = namedtuple("Lookup", ["bind_column", "table", "id"])

"""
Needed because each entity can have several lookups to other entities
"""
EntityLookups = list[Lookup]


class Mapping:
    LOOKUP_TABLE = "new_lookuptable"
    LOOKUP_RELATIONSHIP = "new_lookuprelationship"
    LOOKUP_COLUMN = "new_lookupcolumn"
    DV_COLUMN = "new_dataversecolumnname"
    DV_TABLE = "new_dataversetablenametext"
    FILE_COLUMN = "new_filecolumnname"
    DEFAULT_VALUE = "new_defaultvalue"
    DATA_TYPE = "new_dataversedatatype"
    IS_LOOKUP = "new_islookup"
    IS_KEY = "new_iskey"
    IS_CHOICE = "new_ischoice"
    CHOICE_DICTIONARY = "new_choicedictionary"
    CONDITION = "new_condition"


class Operator(Enum):
    EQUALS = "EQUALS"
    NOT_EQUALS = "NOT_EQUALS"
    IS_NULL = "IS_NULL"
    IS_NOT_NULL = "IS_NOT_NULL"
    SMALLER = "SMALLER"
    GREATER = "GREATER"
    GREATER_OR_EQUAL = "GREATER_OR_EQUAL"
    SMALLER_OR_EQUAL = "SMALLER_OR_EQUAL"


@dataclass
class Condition:
    column: str
    operator: Operator
    is_static_value: bool
    compare_value: Any
    result_value: Any

    @classmethod
    def create(cls, obj: dict) -> "Condition":
        return cls(
            column=obj["column"],
            operator=Operator(obj["operator"]),
            is_static_value=obj["is_static_value"],
            compare_value=obj["compare_value"],
            result_value=obj["result_value"],
        )
    

DEFAULT_SHEET_CHUNK_SIZE = 1000